from burro.processes.discrete_point_process import DiscretePointProcess
from burro.util.sampling import exp_sample


class PoissonProcess(DiscretePointProcess):
    """This class models a discrete poisson process. It uses
    the rate parameter and saves internally the state.
    """

    def __init__(self, lamb):
        """Initialize the process. Simply save the lambda internally."""
        self.lamb = lamb
        self.current_step = 0
        self.overhang_time = 0

    def next(self):
        """This method gets the number of events
        which occured between the timestep the method
        was last called, and the timestep lying steps
        in the future."""

        # precondition
        n = int(self.current_step > 0)

        # loop body
        while True:

            # generate a exponential sample
            y = exp_sample(lamb=self.lamb)
            self.overhang_time += y

            # whenever we reached the limiting timestep
            # break
            if self.overhang_time > 1:
                self.overhang_time -= 1
                break

            # count up by one
            n += 1

        print("poisson generated ", n)
        return n


from burro.agents.agent import Agent


class RefillAgent(Agent):

    def __init__(self, amount, limit):
        self.amount = amount
        self.limit = limit

    def experience(self, state, action, reward, nxt_state, done): pass

    def train(self): pass

    def sync(self): pass

    def act(self, state):
        supply, orders, transported = list(state)
        return self.amount if supply < self.limit else 0



def create_gradient(N, red, green, blue):
    """
    This method creates a gradient by using the supplied lists
    :param red: Red [from, to].
    :param green: Green [from, to].
    :param blue: Blue [from, to]
    :return: A list of [r,g,b] tuples.
    """

    # create color array
    def lin_func(x, lim):
        return (lim[0] + x * (lim[1] - lim[0])) / 255

    def colors(n):
        x = n / (N + 1)
        return lin_func(x, red), lin_func(x, green), lin_func(x, blue)

    return [(0.27, 0.5, 0.71)] + [colors(k) for k in range(N)] + [(0.8, 0.4, 0.4)]



import torch
import torch.nn.functional as F
from torch import distributions
from torch import nn
from typing import Tuple

from burro import BasePolicyModel
from burro import MLP


class GumbelPolicy(torch.nn.Module):

    def __init__(self, in_dim: int, out_dim: int, hidden_size: Tuple[int] = (64, 64,), nonlin: str = 'relu'):
        super(GumbelPolicy, self).__init__()
        self.net = MLP(in_dim, out_dim, hidden_size, nonlin)
        self.temperature = nn.Parameter(torch.ones(1, 1)) # * torch.tensor(std).log())

    def forward(self, x):
        logits = self.net(x)

        return logits, self.temperature


class CategoricalPolicyModel(BasePolicyModel):

    def __init__(self, io_order_dim: int = 3, max_order_size: int = 10, p_hidden_size: Tuple[int] = (64, 64,),
                 vf_hidden_size: Tuple[int] = (64, 64,), p_lr: float = 1e-3, vf_lr: float = 1e-3):
        super(CategoricalPolicyModel, self).__init__()
        self.max_order_size = max_order_size
        self.policy_net = MLP(in_dim=io_order_dim, out_dim=max_order_size, hidden_size=p_hidden_size, nonlin='elu')
        self.vf_net = MLP(in_dim=io_order_dim, out_dim=1, hidden_size=vf_hidden_size, nonlin='elu')
        self.policy_optim = torch.optim.Adam(self.policy_net.parameters(), lr=p_lr)
        self.vf_optim = torch.optim.Adam(self.vf_net.parameters(), lr=vf_lr)

    def act(self, state):
        logits = self.policy_net(state)
        out_orders = distributions.Categorical(logits=logits).sample()
        return out_orders.detach().numpy()

    def sample_gumbel(self, shape, eps=1e-20):
        U = torch.rand(shape)
        return -torch.log(-torch.log(U + eps) + eps)

    def gumbel_softmax_sample(self, logits, temperature):
        y = logits + self.sample_gumbel(logits.size())
        return F.softmax(y / temperature, dim=-1)

    def gumbel_softmax(self, logits, temperature):
        """
        input: [*, n_class]
        return: [*, n_class] an one-hot vector
        """
        y = self.gumbel_softmax_sample(logits, temperature)
        shape = y.size()
        _, ind = y.max(dim=-1)
        y_hard = torch.zeros_like(y).view(-1, shape[-1])
        y_hard.scatter_(1, ind.view(-1, 1), 1)
        y_hard = y_hard.view(*shape)
        return (y_hard - y).detach() + y


if __name__ == '__main__':
    lol = CategoricalPolicyModel()

    x = torch.tensor([5., 4., 2.])

    a = lol.act(x)
    print(a)


import scipy.signal


def discount(x, gamma):
    """
    computes discounted sums along 0th dimension of x.

    inputs
    ------
    x: ndarray
    gamma: float

    outputs
    -------
    y: ndarray with same shape as x, satisfying

        y[t] = x[t] + gamma*x[t+1] + gamma^2*x[t+2] + ... + gamma^k x[t+k],
                where k = len(x) - t - 1

    """
    assert x.ndim >= 1
    return scipy.signal.lfilter([1],[1,-gamma],x[::-1], axis=0)[::-1]


import numpy as np

from burro.processes.discrete_point_process import DiscretePointProcess
from burro.util.type_checking import is_lambda


class HawkesProcess(DiscretePointProcess):

    """Types of hawkes processes."""
    HOMOGENEOUS = 1
    INHOMOGENEOUS = 2

    def __init__(self, intensity):
        """Creates a new hawkes process.

        :param intensity: Can be a float or a function f : R+ -> R+.
        """

        # encapsulate intensity in lambda
        if isinstance(intensity, float):
            self.type = HawkesProcess.HOMOGENEOUS
            self.intensity = lambda t: intensity

        elif is_lambda(intensity):
            self.type = HawkesProcess.INHOMOGENEOUS
            self.intensity = intensity

        self.arrivals = []
        self.eps = 1e-10
        self.t = 0

    def next(self):
        """Get number of arrivals for next timestep

        :return: The number of arrivals
        """

        t = self.t
        self.t += 1
        i = len(self.arrivals)

        while t < self.t:
            m = self.intensity(t + self.eps)
            e = np.exp(m)
            t = t + e
            u = np.random.uniform(0, m)

            if u < self.intensity(t):
                if t < self.t: i += 1
                self.arrivals.append(t)

        self.arrivals = self.arrivals[i:]
        print("hawkes generated ", i)
        return i



import numpy as np

def run(game, agent_factory, n_players, n_steps, render=False, n_sync_steps=10):

    agents = [agent_factory(i) for i in range(n_players)]
    observations = game.reset()

    # repeat
    for k in range(n_steps):

        # obtain new action from each agent
        actions = np.asarray([agents[i].act(observations[i]) for i in range(n_players)])
        nxt_observations, rewards, done = game.step(actions)
        if render: game.render()

        # add experience to each agent and train
        for i in range(n_players):
            agent = agents[i]
            agent.experience(observations[i], actions[i], rewards[i], nxt_observations[i], done)

            # train agent
            agent.train()
            if k % n_sync_steps:
                agent.sync()


import torch
import torch.distributions as dist


def unif_sample(sample_shape=[1], low=0, high=1):
    """Generate a sample from U(low, high).

    :param sample_shape: The shape S
    :param low: Lower bound on uniform
    :param high: Upper bound on uniform
    :return: A matrix of shape S containg samples from the uniform.
    """
    d = dist.Uniform(low, high)
    return d.rsample(sample_shape)


def exp_sample(sample_shape=[1], lamb=1.0):
    """This method generates a sample from an exponential distribution.

    :param sample_shape: The shape S
    :param lamb: The rate of the exponential
    :return: A matrix of shape S containing samples from the exponential.
    """

    u = unif_sample(sample_shape)
    x = torch.log(1 - u) / lamb
    return -x


def gumbel_sample(sample_shape=[1], mu=0.0, beta=1.0):
    """Sample from the gumbel distribution by using the inverse transformation method.

    :param sample_shape: The shape S
    :param mu: The location of the gumbel.
    :param beta: The shape of the gumbel.
    :return: A matrix of shape S containg samples from the gumbel.
    """
    # check if params are valid
    assert beta > 0

    # create distribution and then sample from it

    uniform_samples = unif_sample(sample_shape)
    return mu - beta * torch.log(-torch.log(uniform_samples))


def categorical_sample(p, dim=-1):
    """Perform a categorical sample by using the gumbel max trick.
    See (https://arxiv.org/abs/1611.01144).

    :param p: The distribution probabilities.
    :param dim: The dimension over which the probabiites sum to one
    :return:The sampled actions.
    """

    z = gumbel_sample(p.size())
    return torch.argmax(p + z, dim=dim)



import numpy as np
import matplotlib.pyplot as plt

from burro.processes.hawkes_process import HawkesProcess
from burro.games.markov_game import MarkovGame
from burro.processes.poisson_process import PoissonProcess
from burro.util.colors import create_gradient


class BeerGame(MarkovGame):

    def _state_emission(self, state, i):

        supply, orders, transported = state
        return np.stack([supply[i], transported[i], orders[i+1]])

    def observation_dim(self, i):
        return 3

    def _reset(self):

        return np.zeros(self.num_players), \
               np.zeros(self.num_players+1), \
               np.zeros(self.num_players+1)

    def _state_transition(self, state, actions):

        supply, orders, transported = state

        orders[-1] += self.consumer_process.next()
        orders[:-1] += actions

        transported[0] = orders[0]
        transported[1:] = np.minimum(orders[1:], supply)

        # subtract the number of
        supply += transported[:-1]
        orders -= transported

        cost = self.cost_storage * supply + self.cost_delay * orders[1:]
        supply -= transported[1:]

        # save cost and so on
        self.costs += cost
        self.generated_costs = cost
        self.generated_delay_costs = self.cost_delay * orders[1:]

        return (supply, orders, transported), -cost, False

    def render(self):

        if not self.rendered:

            # init the plot
            self.fig, self.ax = plt.subplots(figsize=(16, 8))
            plt.ion()
            plt.show()
            self.rendered = True

        plt.clf()
        ax = plt.axes()
        golden_ratio = self.gr
        golden_ratio_w = self.gr_w
        left = self.left
        bottom = self.bottom
        box_height = self.box_height
        box_width = self.box_width
        fs = self.fs
        N = self.num_players

        # add background
        v_grey = 0.85
        rect = plt.Rectangle((0, 0), 1, 0.3, facecolor=[v_grey] * 3)
        ax.add_patch(rect)
        rect = plt.Rectangle((0, 0.7), 1, 0.3, facecolor=[v_grey] * 3)
        ax.add_patch(rect)

        # get inner state
        supply, orders, transported = self.inner_state

        for k in range(N + 2):

            args = {}
            if 0 < k < N + 1:
                if self.generated_delay_costs[k - 1] > 0:
                    args['ec'] = 'darkred'

            rect = plt.Rectangle((left[k], bottom), box_width, box_height, facecolor=self.c_colors[k], **args)
            ax.add_patch(rect)

            if k < N + 1:
                arrow_product = plt.Arrow(left[k + 1] - golden_ratio_w,
                                          bottom + 2 * box_height / (2 + golden_ratio),
                                          golden_ratio_w, 0, color="#000000", width=0.01)
                ax.add_patch(arrow_product)

                arrow_order = plt.Arrow(left[k + 1], bottom + box_height / (2 + golden_ratio), -golden_ratio_w, 0,
                                        color="#555555", width=0.01)
                ax.add_patch(arrow_order)
                plt.text(left[k] + box_width + box_width * 0.5 * golden_ratio, bottom + 1.5 * box_height,
                         str(transported[k]),
                         fontsize=fs * golden_ratio, ha='center', va='bottom')

                plt.text(left[k] + box_width + box_width * 0.5 * golden_ratio, bottom - 0.5 * box_height,
                         str(orders[k]),
                         fontsize=fs * golden_ratio, ha='center', va='top')

            if k == 0:
                plt.text(left[k] + 0.5 * box_width, 0.9, "Gen:",
                         fontsize=fs * golden_ratio,
                         color='black', ha='center', va='center')
                plt.text(left[k] + 0.5 * box_width, 0.8, "Sum:", fontsize=fs * golden_ratio,
                         color='black', ha='center', va='center')
                plt.text(left[k] + 0.5 * box_width, bottom + 0.5 * box_height, "∞",
                         fontsize=fs, ha='center', va='center')

            if 0 < k < N + 1:
                plt.text(left[k] + 0.5 * box_width, bottom + 0.5 * box_height, str(supply[k - 1]),
                         fontsize=fs, ha='center', va='center')
                plt.text(left[k] + 0.5 * box_width, 0.9, str(self.generated_costs[k - 1]) + "€",
                         fontsize=fs * golden_ratio,
                         color='darkred', ha='center', va='center')
                plt.text(left[k] + 0.5 * box_width, 0.8, str(self.costs[k - 1]) + "€", fontsize=fs * golden_ratio,
                         color='darkred', ha='center', va='center')

        plt.title("SCM-SIM [" + str(self.current_step) + "]")
        plt.text(0.5, 0.15,
                 "Costs: " + str(np.sum(self.costs)) + "€",
                 fontsize=fs, color='darkred', ha='center', va='center')

        plt.axhline(y=0.7, linestyle='--', color='black')
        plt.axhline(y=0.3, linestyle='--', color='black')
        plt.xlim([0, 1])
        plt.ylim([0, 1])
        plt.xticks([])
        plt.yticks([])
        plt.pause(0.0001)

    plt.show()

    def __init__(self, chain_length, intensity, cost_storage=0.5, cost_delay=1.0, process='poisson'):

        super().__init__(chain_length)

        # create consumer process
        if process == 'poisson': self.consumer_process = PoissonProcess(intensity)
        elif process == 'hawkes': self.hawkes_process = HawkesProcess(intensity)

        self.cost_storage = cost_storage
        self.cost_delay = cost_delay
        self.last_cost = 0
        self.rendered = False
        self.costs = np.zeros(chain_length)

        # calc some drawing related things
        self.fig = None
        self.ax = None

        red = [111, 195]
        green = [137, 225]
        blue = [117, 169]
        self.c_colors = create_gradient(self.num_players, red, green, blue)

        self.gr = 1.64
        self.gr = 1 / self.gr
        h_marg_multiple = self.num_players

        self.width = self.num_players + (self.num_players + 3) * self.gr + 2
        self.height = 1 + 2 * h_marg_multiple * self.gr
        self.gr_w = self.gr / self.width
        self.gr_h = self.gr / self.height

        self.left = np.arange(self.num_players + 2) * (1 / self.width + self.gr_w) + self.gr_w
        self.bottom = self.num_players * self.gr_h
        self.box_width = 1 / self.width
        self.box_height = 1 / self.height
        self.fs = 245 * self.box_height



import torch
from torch import distributions
from copy import deepcopy

from burro.agents import Agent
from burro import CategoricalPolicyModel
from burro import discount


class CategoricalAgent(Agent):

    def __init__(self, policy_model: CategoricalPolicyModel, gamma: float = 0.99,
                 batch_size: int = 100, vf_epochs: int = 25):
        super(CategoricalAgent, self).__init__()
        self.policy_model = policy_model
        self.max_order_size = self.policy_model.max_order_size
        self.batch_size = batch_size
        self.gamma = gamma
        self.states_batch = torch.empty((self.batch_size, 3))
        self.reward_batch = torch.zeros((self.batch_size, 1))
        self.oorders_batch = torch.empty((self.batch_size, 1))
        self.batch_fill_count = 0
        self.vf_epochs = vf_epochs

    def clone(self):
        return CategoricalAgent(deepcopy(self.policy_model), gamma=self.gamma, batch_size=self.batch_size,
                                vf_epochs=self.vf_epochs)

    def get_outgoing_orders(self, pos, clen, stock, iorders, oorders, costs):
        state = torch.stack([torch.tensor(stock, dtype=torch.float32), torch.tensor(iorders, dtype=torch.float32),
                             torch.tensor(oorders, dtype=torch.float32)])
        new_oorders = self.policy_model.act(state)
        self.oorders_batch[self.batch_fill_count] = torch.from_numpy(new_oorders)
        self.states_batch[self.batch_fill_count] = state
        self.reward_batch[self.batch_fill_count] = -costs # (torch.eye(self.max_order_size)[iorders] - train_orders).mean()
        self.batch_fill_count += 1
        if self.batch_fill_count >= self.batch_size:
            discounted_reward = discount(self.reward_batch.numpy(), self.gamma)
            discounted_reward = torch.cat([torch.tensor(dr) for dr in discounted_reward])
            self.batch_fill_count = 0
            self.update_vf(self.states_batch, discounted_reward)
            advantage = discounted_reward - self.policy_model.vf_net(self.states_batch)
            self.update_policy(self.states_batch, advantage, self.oorders_batch)

        return new_oorders

    def update_policy(self, states, advantage, oorders):
        logits = self.policy_model.policy_net(states)
        log_probs = torch.distributions.Categorical(logits=logits).log_prob(oorders)
        loss = - (advantage * log_probs).mean()
        self.policy_model.policy_optim.zero_grad()
        loss.backward()
        self.policy_model.policy_optim.step()

    def update_vf(self, states, reward):
        loss_fn = torch.nn.MSELoss()
        for _ in range(self.vf_epochs):
            values = self.policy_model.vf_net(states)
            loss = loss_fn(reward[:, None], values).mean(0)
            self.policy_model.vf_optim.zero_grad()
            loss.backward()
            self.policy_model.vf_optim.step()



from abc import ABC, abstractmethod


class Agent(ABC):
    """Represents an interface for one agent. Different implementations
    can be used as this will be used inside the simulation to drive the
    different goods."""

    @abstractmethod
    def experience(self, state, action, reward, nxt_state, done): pass

    @abstractmethod
    def train(self): pass

    @abstractmethod
    def sync(self): pass

    @abstractmethod
    def act(self, state):
        """This method should calculate outgoing orders. Therefore several information is
        about itself is available.

        :param pos: The position in the product flow, where 0 means first node.
        :param clen: The length of the chain, for getting position in order flow.
        :param stock: The supply available in the stock.
        :param iorders: The deliveries requested by the node (pos + 1)
        :param oorders: The deliveries requested by you in the previous step.
        :param costs: The costs produced in last step.

        :return: How much should be ordered from (pos - 1)
        """
        pass



import setuptools
from setuptools import setup

with open('requirements.txt') as f:
    required = f.read().splitlines()

import numpy as np

setup(name='burrolib',
      version='0.1.1',
      description='burro - a supply chain simulator',
      author='xploras',
      author_email='dev@xploras.net',
      url="https://github.com/kosmitive/burro",
      install_requires=required,
      packages=setuptools.find_packages(),
      include_dirs=[np.get_include()],
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
      ],
      python_requires='>=3.7',
      )



import torch
from typing import Tuple


class MLP(torch.nn.Module):

    def __init__(self, in_dim: int, out_dim: int, hidden_size: Tuple[int] = (64,), nonlin: str = 'relu', **kwargs):
        super(MLP, self).__init__()
        nlist = dict(relu=torch.nn.ReLU(), tanh=torch.nn.Tanh(),
                     sigmoid=torch.nn.Sigmoid(), softplus=torch.nn.Softplus(), lrelu=torch.nn.LeakyReLU(),
                     elu=torch.nn.ELU())

        self.nonlin = nlist[nonlin]
        self.net = torch.nn.ModuleList([torch.nn.Linear(in_dim, hidden_size[0]),
                                  self.nonlin])
        if len(hidden_size) > 1:
            for i in range(len(hidden_size) - 1):
                self.net.append(torch.nn.Linear(hidden_size[i], hidden_size[i + 1]))
        self.out = torch.nn.Linear(hidden_size[-1], out_dim)

    def forward(self, x, **kwargs):
        for _, layer in enumerate(self.net):
            x = layer(x)

        return self.out(x)



class MarkovGame:

    def __init__(self, num_players):
        """

        :param num_players: Define the number of different players
        :param s0: The state transition function.
        """

        # save general settings
        self.num_players = num_players
        self.inner_state = self.reset()
        self.current_step = 0

    def step(self, actions):
        self.inner_state, rewards, done = self._state_transition(self.inner_state, actions)
        nxt_observations = self.get_observations()
        self.current_step += 1
        return nxt_observations, rewards, done

    def observation_dim(self, i):
        raise NotImplementedError

    def _state_transition(self, state, actions):
        raise NotImplementedError

    def _state_emission(self, state, i):
        raise NotImplementedError

    def _reset(self):
        raise NotImplementedError

    def render(self):
        raise NotImplementedError

    def reset(self):
        self.inner_state = self._reset()
        return self.get_observations()

    def get_observations(self):
        return [self._state_emission(self.inner_state, i) for i in range(self.num_players)]



import torch
import numpy as np


def tile(a, dim, n_tile):
    """TIle the input array along the dimension.

    :param a: The tensor to repeat.
    :param dim: The dimension to repeat.
    :param n_tile: The number of times the dimension should be repeated.
    :return: The tiles torch tensor.
    """

    init_dim = a.size(dim)
    repeat_idx = [1] * a.dim()
    repeat_idx[dim] = n_tile
    a = a.repeat(*repeat_idx)
    order_index = torch.LongTensor(np.concatenate([init_dim * np.arange(n_tile) + i for i in range(init_dim)]))
    return torch.index_select(a, dim, order_index)
